import math

def euclidean_distance(point1, point2):
  '''
  Args:
    point1: type array two values [x, y]
    point2: type array two values [x, y]
  Returns:
    Distance of two points
  Descriptions:
    Calculate Euclidian distance of two points in Euclidian space
  '''

  return math.sqrt((point1[0]-point2[0])**2 + (point1[1]-point2[1])**2)

def curve_length(curve):
  '''
  Args:
    points: type arrays two values [[x, y], [x, y]]
  Returns:
    acc_length: curve length
  Descriptions:
    Calculate the length of the curve
  '''

  acc_length = 0
  for i in range(0, len(curve)-1):
    acc_length += euclidean_distance(curve[i], curve[i+1])
  return acc_length

def extend_point_on_line(point1, point2, distance):
  '''
  Args:
    point1: type array two values [x, y]
    point2: type array two values [x, y]
    distance: type float
  Returns:
    A new point, point3, which is on the same
    line generated by point1 and point2
  '''

  norm = round(distance / euclidean_distance(point1, point2), 2)
  new_point_x = point2[0] + norm * (point1[0] - point2[0])
  new_point_y = point2[1] + norm * (point1[1] - point2[1])
  return [new_point_x, new_point_y]

def rotate_curve(curve, thetaRad):
  '''
  Args:
    curve: original curve, type array two values [[x, y], [x, y]]
    thetaRad: rotation angle type float
  Returns:
    rot_curve: rotated curve
  Descriptions:
    Rotate the curve around the origin
  '''

  rot_curve = []
  for i in range(len(curve)):
    x_cord = math.cos(-1 * thetaRad) * curve[i][0] - math.sin(-1 * thetaRad) * curve[i][1]
    y_cord = math.sin(-1 * thetaRad) * curve[i][0] + math.cos(-1 * thetaRad) * curve[i][1]
    rot_curve.append([x_cord, y_cord])
  return rot_curve



from .geometry import *
from .utils import *
import math

'''
from https://en.wikipedia.org/wiki/Procrustes_analysis
'''

def procrustes_normalize_curve(curve):
  '''
  Args:
    curve: type array [[x, y]], [x, y]].
  Returns:
    procrustes_normalize_curve: procrustes_normalize_curve
  Descriptions:
    Translate and scale curve by Procrustes Analysis
  '''
  
  curve_length = len(curve)
  mean_x = array_average(list(map(lambda item: item[0], curve)))
  mean_y = array_average(list(map(lambda item: item[1], curve)))
  curve = list(map(lambda item: [item[0]-mean_x, item[1]-mean_y], curve))
  
  squared_sum = 0
  for i in range(curve_length):
    squared_sum += (curve[i][0])**2 + (curve[i][1])**2
  scale = round(squared_sum / curve_length, 2)
  return list(map(lambda item: [item[0]/scale, item[1]/scale], curve))

def find_procrustes_rotation_angle(curve, relativeCurve):
  '''
  Args:
    curve: type array [[x, y]], [x, y]].
    relativeCurve: type array [[x, y]], [x, y]].
  Warnings:
    `curve` and `relativeCurve` must have the same number of points
    `curve` and `relativeCurve` should both be run through [[procrustesNormalizeCurve]] first
  Returns:
    find_procrustes_rotation_angle: the angle to rotate
  Descriptions:
    Find the angle to rotate `curve` to match the rotation
    of `relativeCurve` using procrustes analysis
  '''
  
  assert len(curve) == len(relativeCurve), 'curve and relativeCurve must have the same length'
  numerator, denominator = 0, 0
  for i in range(len(curve)):
    numerator += relativeCurve[i][0]*curve[i][1] - relativeCurve[i][1]*curve[i][0]
    denominator += relativeCurve[i][0]*curve[i][0] + relativeCurve[i][1]*curve[i][1]
  return math.atan2(numerator, denominator)

def procrustes_normalize_rotation(curve, relativeCurve):
  '''
  Args:
    curve: type array [[x, y]], [x, y]].
    relativeCurve: type array [[x, y]], [x, y]].
  Warnings:
    `curve` and `relativeCurve` must have the same number of points
    `curve` and `relativeCurve` should both be run through [[procrustesNormalizeCurve]] first
  Returns:
    procrustes_normalize_rotation: rotate
  Descriptions:
    Rotate `curve` to match the rotation of 
    `relativeCurve` using procrustes analysis
  '''

  angle = find_procrustes_rotation_angle(curve, relativeCurve)
  return rotate_curve(curve, angle)


from .procrustesanalysis import *
from .frechetdistance import *
from .geometry import *
import math

def shape_similarity(shape1, shape2, rotations=10, checkRotation=True):
  procrustes_normalized_curve1 = procrustes_normalize_curve(shape1)
  procrustes_normalized_curve2 = procrustes_normalize_curve(shape2)
  geo_avg_curve_len = math.sqrt(
    curve_length(procrustes_normalized_curve1) * 
    curve_length(procrustes_normalized_curve2)
  )

  thetas_to_check = [0]
  if checkRotation:
    procrustes_theta = find_procrustes_rotation_angle(
      procrustes_normalized_curve1, 
      procrustes_normalized_curve2
    )
    # use a negative rotation rather than a large positive rotation
    if procrustes_theta > math.pi:
      procrustes_theta = procrustes_theta - 2 * math.pi
    if procrustes_theta != 0 and abs(procrustes_theta) < math.pi:
      thetas_to_check.append(procrustes_theta)
    for i in range(0, rotations):
      theta = -1 * math.pi + (2 * i * math.pi) / (rotations - 1)
      # 0 and Math.PI are already being checked, no need to check twice
      if theta != 0 and theta != math.pi:
        thetas_to_check.append(theta)

  # Using Frechet distance to check the similarity level 
  min_frechet_distance = float('inf')
  # check some other thetas here just in case the procrustes theta isn't the best rotation
  for theta in thetas_to_check:
    rotated_curve1 = rotate_curve(procrustes_normalized_curve1, theta)
    frechet_dist = frechet_distance(rotated_curve1, procrustes_normalized_curve2)
    if frechet_dist < min_frechet_distance:
      min_frechet_distance = frechet_dist
  # divide by Math.sqrt(2) to try to get the low results closer to 
  result = max(1 - min_frechet_distance / (geo_avg_curve_len / math.sqrt(2)), 0)
  return round(result, 4)


from setuptools import setup
import io

setup(
  name='shapesimilarity',
  version='1.0.0',
  author='Nelson Wenner',
  packages=['shapesimilarity'],
  url='https://github.com/nelsonwenner/shape-similarity',
  description='Quantify the similarity between two shapes/curves',
  long_description_content_type="text/markdown",
  long_description=io.open('README.md', encoding="utf-8").read(),
  keywords=['python', 'curve', 'frechet-distance', 'procrustes-analysis'],
  platforms=['any']
)
